import{b as d,C as l,r as n,a as p,t as i,c as g,f as S,d as u,u as E,s as A,R as f,j as s,P as h,A as m,U as P,e as _}from"./appReducer-94wa7pMY.js";console.debug("baseUrl:",d);l.load().then(()=>{const r=(x,a)=>a.type!==P&&a.type!==_,o=n.createLogger({collapsed:!0,diff:!1,predicate:r}),c=p(i,o),e=g(S,c),t=e.dispatch;t(u(e.getState().controlState.locale)),t(E()),e.getState().controlState.privacyNoticeAccepted&&t(A(e,!0)),f.createRoot(document.getElementById("root")).render(s.jsx(h,{store:e,children:s.jsx(m,{})}))});
