# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .context import PlacesContext
from .routes import PATH_PARAM_PLACE_GROUP_ID, api
