import{b as n,C as l,r as d,a as u,c as p,d as f,u as g,s as m,R as y,j as i,P as h,A as S,U as L,e as P,f as E,t as A}from"./appReducer-94wa7pMY.js";(function(){const s=document.createElement("link").relList;if(s&&s.supports&&s.supports("modulepreload"))return;for(const e of document.querySelectorAll('link[rel="modulepreload"]'))r(e);new MutationObserver(e=>{for(const t of e)if(t.type==="childList")for(const o of t.addedNodes)o.tagName==="LINK"&&o.rel==="modulepreload"&&r(o)}).observe(document,{childList:!0,subtree:!0});function a(e){const t={};return e.integrity&&(t.integrity=e.integrity),e.referrerPolicy&&(t.referrerPolicy=e.referrerPolicy),e.crossOrigin==="use-credentials"?t.credentials="include":e.crossOrigin==="anonymous"?t.credentials="omit":t.credentials="same-origin",t}function r(e){if(e.ep)return;e.ep=!0;const t=a(e);fetch(e.href,t)}})();console.debug("baseUrl:",n);l.load().then(()=>{const c=(t,o)=>o.type!==L&&o.type!==P,s=d.createLogger({collapsed:!0,diff:!1,predicate:c}),a=u(A,s),r=p(E,a),e=r.dispatch;e(f(r.getState().controlState.locale)),e(g()),r.getState().controlState.privacyNoticeAccepted&&e(m(r,!0)),y.createRoot(document.getElementById("root")).render(i.jsx(h,{store:r,children:i.jsx(S,{})}))});
