# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .routes import (
    PATH_PARAM_X,
    PATH_PARAM_Y,
    PATH_PARAM_Z,
    QUERY_PARAM_FORMAT,
    QUERY_PARAM_RETINA,
    QUERY_PARAM_TIME,
    api,
)
